# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import fontTools.pens.boundsPen
from fontTools.misc.transform import Transform

from ufo2ft.filters import BaseFilter, BaseIFilter
from ufo2ft.util import OpenTypeCategories, zip_strict

logger = logging.getLogger(__name__)


class PropagateAnchorsFilter(BaseFilter):
    def set_context(self, font, glyphSet):
        ctx = super().set_context(font, glyphSet)
        ctx.processed = set()
        ctx.categories = OpenTypeCategories.load(font)
        return ctx

    def __call__(self, font, glyphSet=None):
        modified = super().__call__(font, glyphSet)
        if modified:
            logger.info("Glyphs with propagated anchors: %i" % len(modified))
        return modified

    def filter(self, glyph):
        if not glyph.components:
            return False
        before = len(glyph.anchors)
        _propagate_glyph_anchors(
            self.context.glyphSet,
            glyph,
            self.context.processed,
            self.context.modified,
            self.context.categories,
        )
        return len(glyph.anchors) > before


class PropagateAnchorsIFilter(BaseIFilter):
    def set_context(self, *args, **kwargs):
        ctx = super().set_context(*args, **kwargs)
        ctx.processed = [set() for _ in range(len(ctx.glyphSets))]
        ctx.categories = OpenTypeCategories.load(self.getDefaultFont())
        return ctx

    def __call__(self, fonts, glyphSets=None, instantiator=None, **kwargs):
        modified = super().__call__(fonts, glyphSets, instantiator, **kwargs)
        if modified:
            logger.info("Glyphs with propagated anchors: %i" % len(modified))
        return modified

    def filter(self, glyphName, glyphs):
        modified = False
        if not any(glyph.components for glyph in glyphs):
            return modified
        before = len(self.context.modified)
        for i, (glyphSet, interpolatedLayer) in enumerate(
            zip_strict(self.context.glyphSets, self.getInterpolatedLayers())
        ):
            glyph = glyphSet.get(glyphName)
            if glyph is not None:
                _propagate_glyph_anchors(
                    interpolatedLayer or glyphSet,
                    glyph,
                    self.context.processed[i],
                    self.context.modified,
                    self.context.categories,
                )
        return len(self.context.modified) > before


def _propagate_glyph_anchors(glyphSet, composite, processed, modified, categories):
    """
    Propagate anchors from base glyphs to a given composite
    glyph, and to all composite glyphs used in between.
    """

    if composite.name in processed:
        return
    processed.add(composite.name)

    if not composite.components or (
        # "If it is a 'mark' and there are anchors, it will not look into components"
        # Georg said: https://github.com/googlefonts/ufo2ft/issues/802#issuecomment-1904109457
        composite.name in categories.mark
        and composite.anchors
    ):
        return

    base_components = []
    mark_components = []
    anchor_names = set()
    to_add = {}
    for component in composite.components:
        try:
            glyph = glyphSet[component.baseGlyph]
        except KeyError:
            logger.warning(
                "Anchors not propagated for inexistent component {} "
                "in glyph {}".format(component.baseGlyph, composite.name)
            )
        else:
            _propagate_glyph_anchors(glyphSet, glyph, processed, modified, categories)
            if any(a.name.startswith("_") for a in glyph.anchors):
                mark_components.append(component)
            else:
                base_components.append(component)
                anchor_names |= {a.name for a in glyph.anchors}

    if mark_components and not base_components and _is_ligature_mark(composite):
        # The composite is a mark that is composed of other marks (E.g.
        # "circumflexcomb_tildecomb"). Promote the mark that is positioned closest
        # to the origin to a base.
        try:
            component = _component_closest_to_origin(mark_components, glyphSet)
        except Exception as e:
            raise Exception(
                "Error while determining which component of composite "
                "'{}' is the lowest: {}".format(composite.name, str(e))
            ) from e
        mark_components.remove(component)
        base_components.append(component)
        glyph = glyphSet[component.baseGlyph]
        anchor_names |= {a.name for a in glyph.anchors}

    for anchor_name in anchor_names:
        # don't add if composite glyph already contains this anchor OR any
        # associated ligature anchors (e.g. "top_1, top_2" for "top")
        if not any(a.name.startswith(anchor_name) for a in composite.anchors):
            _get_anchor_data(to_add, glyphSet, base_components, anchor_name)

    for component in mark_components:
        _adjust_anchors(to_add, glyphSet, component)

    # we sort propagated anchors to append in a deterministic order
    for name, (x, y) in sorted(to_add.items()):
        anchor_dict = {"name": name, "x": x, "y": y}
        try:
            composite.appendAnchor(anchor_dict)
        except TypeError:  # pragma: no cover
            # fontParts API
            composite.appendAnchor(name, (x, y))

    if to_add:
        modified.add(composite.name)


def _get_anchor_data(anchor_data, glyphSet, components, anchor_name):
    """Get data for an anchor from a list of components."""

    anchors = []
    for component in components:
        for anchor in glyphSet[component.baseGlyph].anchors:
            if anchor.name == anchor_name:
                anchors.append((anchor, component))
                break
    if len(anchors) > 1:
        for i, (anchor, component) in enumerate(anchors):
            t = Transform(*component.transformation)
            name = "%s_%d" % (anchor.name, i + 1)
            anchor_data[name] = t.transformPoint((anchor.x, anchor.y))
    elif anchors:
        anchor, component = anchors[0]
        t = Transform(*component.transformation)
        anchor_data[anchor.name] = t.transformPoint((anchor.x, anchor.y))


def _adjust_anchors(anchor_data, glyphSet, component):
    """
    Adjust base anchors to which a mark component may have been attached, by
    moving the base anchor attached to a mark anchor to the position of
    the mark component's base anchor.
    """

    glyph = glyphSet[component.baseGlyph]
    t = Transform(*component.transformation)
    for anchor in glyph.anchors:
        # only adjust if this anchor has data and the component also contains
        # the associated mark anchor (e.g. "_top" for "top")
        if anchor.name in anchor_data and any(
            a.name == "_" + anchor.name for a in glyph.anchors
        ):
            anchor_data[anchor.name] = t.transformPoint((anchor.x, anchor.y))


def _component_closest_to_origin(components, glyph_set):
    """Return the component whose (xmin, ymin) bounds are closest to origin.

    This ensures that a component that is moved below another is
    actually recognized as such. Looking only at the transformation
    offset can be misleading.
    """
    return min(components, key=lambda comp: _distance((0, 0), _bounds(comp, glyph_set)))


def _distance(pos1, pos2):
    x1, y1 = pos1
    x2, y2 = pos2
    return (x1 - x2) ** 2 + (y1 - y2) ** 2


def _is_ligature_mark(glyph):
    return not glyph.name.startswith("_") and "_" in glyph.name


def _bounds(component, glyph_set):
    """Return the (xmin, ymin) of the bounds of `component`."""
    if hasattr(component, "bounds"):  # e.g. defcon
        return component.bounds[:2]
    elif hasattr(component, "draw"):  # e.g. ufoLib2
        pen = fontTools.pens.boundsPen.BoundsPen(glyphSet=glyph_set)
        component.draw(pen)
        return pen.bounds[:2]
    else:
        raise ValueError(
            f"Don't know to to compute the bounds of component '{component}' "
        )



from fontTools.feaLib.variableScalar import VariableScalar
from fontTools.misc.fixedTools import otRound

from ufo2ft.featureWriters import BaseFeatureWriter, ast


def caretSortKey(caret):
    if isinstance(caret, VariableScalar):
        return list(caret.values.values())[0]
    return caret


class GdefFeatureWriter(BaseFeatureWriter):
    """Generates a GDEF table based on OpenType Category and glyph anchors.

    It skips generating the GDEF if a GDEF is defined in the features.

    It uses the 'public.openTypeCategories' values to create the GDEF ClassDefs
    and the ligature caret anchors to create the GDEF ligature carets.

    """

    tableTag = "GDEF"
    features = frozenset(["GlyphClassDefs", "LigatureCarets"])
    insertFeatureMarker = None

    def setContext(self, font, feaFile, compiler=None):
        ctx = super().setContext(font, feaFile, compiler=compiler)

        ctx.gdefTableBlock = ast.findTable(self.context.feaFile, "GDEF")
        if ctx.gdefTableBlock:
            for fea in ctx.gdefTableBlock.statements:
                if isinstance(fea, ast.GlyphClassDefStatement):
                    ctx.todo.discard("GlyphClassDefs")
                elif isinstance(fea, ast.LigatureCaretByIndexStatement) or isinstance(
                    fea, ast.LigatureCaretByPosStatement
                ):
                    ctx.todo.discard("LigatureCarets")

                if not ctx.todo:
                    break

        ctx.orderedGlyphSet = self.getOrderedGlyphSet()

        if "GlyphClassDefs" in ctx.todo:
            ctx.openTypeCategories = self.getOpenTypeCategories()
            if not any(ctx.openTypeCategories):
                ctx.todo.remove("GlyphClassDefs")
        if "LigatureCarets" in ctx.todo:
            ctx.ligatureCarets = self._getLigatureCarets()
            if not ctx.ligatureCarets:
                ctx.todo.remove("LigatureCarets")

        return ctx

    def _getLigatureCarets(self):
        carets = dict()

        for glyphName, glyph in self.context.orderedGlyphSet.items():
            glyphCarets = set()
            for anchor in glyph.anchors:
                if (
                    anchor.name
                    and anchor.name.startswith("caret_")
                    and anchor.x is not None
                ):
                    glyphCarets.add(self._getAnchor(glyphName, anchor.name)[0])
                elif (
                    anchor.name
                    and anchor.name.startswith("vcaret_")
                    and anchor.y is not None
                ):
                    glyphCarets.add(self._getAnchor(glyphName, anchor.name)[1])

            if glyphCarets:
                if self.context.isVariable:
                    carets[glyphName] = sorted(glyphCarets, key=caretSortKey)
                else:
                    carets[glyphName] = [otRound(c) for c in sorted(glyphCarets)]

        return carets

    def _sortedGlyphClass(self, glyphNames):
        return sorted(n for n in self.context.orderedGlyphSet if n in glyphNames)

    def _write(self):
        feaFile = self.context.feaFile
        gdefTableBlock = self.context.gdefTableBlock
        if not gdefTableBlock:
            gdefTableBlock = ast.TableBlock("GDEF")
            feaFile.statements.append(gdefTableBlock)

        if "GlyphClassDefs" in self.context.todo:
            categories = self.context.openTypeCategories
            glyphClassDefs = ast.GlyphClassDefStatement(
                ast.GlyphClass(self._sortedGlyphClass(categories.base)),
                ast.GlyphClass(self._sortedGlyphClass(categories.mark)),
                ast.GlyphClass(self._sortedGlyphClass(categories.ligature)),
                ast.GlyphClass(self._sortedGlyphClass(categories.component)),
            )
            gdefTableBlock.statements.append(glyphClassDefs)

        if "LigatureCarets" in self.context.todo:
            ligatureCarets = [
                ast.LigatureCaretByPosStatement(ast.GlyphName(glyphName), carets)
                for glyphName, carets in self.context.ligatureCarets.items()
            ]
            gdefTableBlock.statements.extend(ligatureCarets)

        return True



import pytest
from fontTools.ttLib import TTFont

from ufo2ft.infoCompiler import InfoCompiler

from .outlineCompiler_test import getpath


@pytest.fixture
def testttf():
    font = TTFont()
    font.importXML(getpath("TestFont.ttx"))
    return font


@pytest.fixture
def testufo(FontClass):
    font = FontClass(getpath("TestFont.ufo"))
    return font


class InfoCompilerTest:
    def test_head(self, testttf, testufo):
        info = {"versionMajor": 5, "versionMinor": 6}
        compiler = InfoCompiler(testttf, testufo, info)
        ttf = compiler.compile()
        assert ttf["head"].fontRevision == 5.006

    def test_hhea(self, testttf, testufo):
        info = {"openTypeHheaAscender": 100, "openTypeHheaDescender": -200}
        compiler = InfoCompiler(testttf, testufo, info)
        ttf = compiler.compile()
        assert ttf["hhea"].ascent == 100
        assert ttf["hhea"].descent == -200

    def test_vhea(self, testttf, testufo):
        info = {
            "openTypeVheaVertTypoAscender": 100,
            "openTypeVheaVertTypoDescender": -200,
        }
        compiler = InfoCompiler(testttf, testufo, info)
        ttf = compiler.compile()
        assert ttf["vhea"].ascent == 100
        assert ttf["vhea"].descent == -200

    def test_name(self, testttf, testufo):
        info = {"postscriptFontName": "TestFontOverride-Italic"}
        compiler = InfoCompiler(testttf, testufo, info)
        ttf = compiler.compile()
        assert ttf["name"].getDebugName(6) == "TestFontOverride-Italic"

    def test_OS2(self, testttf, testufo):
        info = {"openTypeOS2TypoAscender": 100, "openTypeOS2TypoDescender": -200}
        compiler = InfoCompiler(testttf, testufo, info)
        ttf = compiler.compile()
        assert ttf["OS/2"].sTypoAscender == 100
        assert ttf["OS/2"].sTypoDescender == -200

    def test_post(self, testttf, testufo):
        info = {"italicAngle": 30.6}
        compiler = InfoCompiler(testttf, testufo, info)
        ttf = compiler.compile()
        assert ttf["post"].italicAngle == 30.6

    def test_gasp(self, testttf, testufo):
        info = {
            "openTypeGaspRangeRecords": [
                {
                    "rangeMaxPPEM": 8,
                    "rangeGaspBehavior": [0, 2],
                }
            ]
        }
        compiler = InfoCompiler(testttf, testufo, info)
        ttf = compiler.compile()
        assert ttf["gasp"].gaspRange == {8: 5}



import logging
from collections import OrderedDict
from types import SimpleNamespace

from fontTools.designspaceLib import DesignSpaceDocument
from fontTools.feaLib.variableScalar import VariableScalar
from fontTools.misc.fixedTools import otRound

from ufo2ft.errors import InvalidFeaturesData
from ufo2ft.featureWriters import ast
from ufo2ft.util import (
    OpenTypeCategories,
    collapse_varscalar,
    get_userspace_location,
    quantize,
    unicodeScriptExtensions,
)

INSERT_FEATURE_MARKER = r"\s*# Automatic Code.*"


class BaseFeatureWriter:
    """Abstract features writer.

    The `tableTag` class attribute (str) states the tag of the OpenType
    Layout table which the generated features are intended for.
    For example: "GPOS", "GSUB", "BASE", etc.

    The `features` class attribute defines the set of all the features
    that this writer supports. If you want to only write some of the
    available features you can provide a smaller sequence to 'features'
    constructor argument. By the default all the features supported by
    this writer will be outputted.

    Two writing modes are defined here:
    1) "skip" (default) will not write features if already present;
    2) "append" will add additional lookups to an existing feature,
       if present, or it will add a new one at the end of all features.
    Subclasses can set a different default mode or define a different
    set of `_SUPPORTED_MODES`.

    The `options` class attribute contains a mapping of option
    names with their default values. These can be overridden on an
    instance by passing keyword arguments to the constructor.

    Combining manually written and automatically generated feature code
    can be achieved by using the `# Automatic Code` insertion marker
    inside the feature code. Automatically generated code for the
    respective feature is added in that spot.
    """

    tableTag = None
    features = frozenset()
    mode = "skip"
    insertFeatureMarker = INSERT_FEATURE_MARKER
    options = {}

    _SUPPORTED_MODES = frozenset(["skip", "append"])

    def __init__(self, features=None, mode=None, **kwargs):
        if features is not None:
            features = frozenset(features)
            assert features, "features cannot be empty"
            unsupported = features.difference(self.__class__.features)
            if unsupported:
                raise ValueError("unsupported: %s" % ", ".join(unsupported))
            self.features = features

        if mode is not None:
            self.mode = mode
        if self.mode not in self._SUPPORTED_MODES:
            raise ValueError(self.mode)

        options = dict(self.__class__.options)
        for k in kwargs:
            if k not in options:
                raise TypeError("unsupported keyword argument: %r" % k)
            options[k] = kwargs[k]
        self.options = SimpleNamespace(**options)

        logger = ".".join([self.__class__.__module__, self.__class__.__name__])
        self.log = logging.getLogger(logger)

    def setContext(self, font, feaFile, compiler=None):
        """Populate a temporary `self.context` namespace, which is reset
        after each new call to `_write` method.
        Subclasses can override this to provide contextual information
        which depends on other data, or set any temporary attributes.

        The default implementation sets:
        - the current font;
        - the current FeatureFile object;
        - the current compiler instance (only present when this writer was
          instantiated from a FeatureCompiler);
        - a set of features (tags) to be generated. If self.mode is "skip",
          these are all the features which are _not_ already present.
        - a set of all existing features (tags) in the feaFile.

        Returns the context namespace instance.
        """
        todo = set(self.features)
        insertComments = None
        existing = set()
        if self.mode == "skip":
            if self.insertFeatureMarker is not None:
                insertComments = self.collectInsertMarkers(
                    feaFile, self.insertFeatureMarker, todo
                )
            # find existing feature blocks
            existing = ast.findFeatureTags(feaFile)
            # ignore features with insert marker
            if insertComments:
                existing.difference_update(insertComments.keys())
            # remove existing feature without insert marker from todo list
            todo.difference_update(existing)

        self.context = SimpleNamespace(
            font=font,
            feaFile=feaFile,
            compiler=compiler,
            todo=todo,
            insertComments=insertComments,
            existingFeatures=existing,
            isVariable=isinstance(font, DesignSpaceDocument),
        )

        return self.context

    def shouldContinue(self):
        """Decide whether to start generating features or return early.
        Returns a boolean: True to proceed, False to skip.

        Sublcasses may override this to skip generation based on the presence
        or lack of other required pieces of font data.
        """
        if not self.context.todo:
            self.log.debug("No features to be generated; skipped")
            return False
        return True

    def write(self, font, feaFile, compiler=None):
        """Write features and class definitions for this font to a feaLib
        FeatureFile object.

        The main entry point for the FeatureCompiler to any of the
        FeatureWriters.

        Returns True if feature file was modified, False if no new features were
        generated.
        """
        self.setContext(font, feaFile, compiler=compiler)
        try:
            if self.shouldContinue():
                return self._write()
            else:
                return False
        finally:
            del self.context

    def _write(self):
        """Subclasses must override this."""
        raise NotImplementedError

    def _insert(
        self,
        feaFile,
        classDefs=None,
        anchorDefs=None,
        markClassDefs=None,
        lookups=None,
        features=None,
    ):
        """
        Insert feature, its classDefs or markClassDefs and lookups at insert
        marker comment.

        If the insert marker is at the top of a feature block, the feature is
        inserted before that block, and after if the insert marker is at the
        bottom.
        """

        statements = feaFile.statements
        inserted = {}

        # First handle those with a known location, i.e. insert markers
        insertComments = self.context.insertComments
        indices = []
        for ix, feature in enumerate(features):
            if insertComments and feature.name in insertComments:
                block, comment = insertComments[feature.name]
                markerIndex = block.statements.index(comment)

                onlyCommentsBefore = all(
                    isinstance(s, ast.Comment) for s in block.statements[:markerIndex]
                )
                onlyCommentsAfter = all(
                    isinstance(s, ast.Comment) for s in block.statements[markerIndex:]
                )

                # Remove insert marker(s) from feature block.
                del block.statements[markerIndex]

                # insertFeatureMarker is in a block with only comments.
                # Replace that block with new feature block.
                if onlyCommentsBefore and onlyCommentsAfter:
                    index = statements.index(block)
                    statements.remove(block)

                # insertFeatureMarker is at the top of a feature block
                # or only preceded by other comments.
                elif onlyCommentsBefore:
                    index = statements.index(block)

                # insertFeatureMarker is at the bottom of a feature block
                # or only followed by other comments
                elif onlyCommentsAfter:
                    index = statements.index(block) + 1

                # insertFeatureMarker is in the middle of a feature block
                # preceded and followed by statements that are not comments
                #
                # Glyphs3 can insert a feature block when rules are before
                # and after the insert marker.
                # See
                # https://github.com/googlefonts/ufo2ft/issues/351#issuecomment-765294436
                # This is currently not supported.
                else:
                    raise InvalidFeaturesData(
                        "Insert marker has rules before and after, feature "
                        f"{block.name} cannot be inserted. This is not supported."
                    )

                statements.insert(index, feature)
                indices.append(index)
                inserted[id(feature)] = True

                # Now walk feature list backwards and insert any dependent features
                for i in range(ix - 1, -1, -1):
                    if id(features[i]) in inserted:
                        break
                    # Insert this before the current one i.e. at same array index
                    statements.insert(index, features[i])
                    # All the indices recorded previously have now shifted up by one
                    indices = [index] + [j + 1 for j in indices]
                    inserted[id(features[i])] = True

        # Finally, deal with any remaining features
        for feature in features:
            if id(feature) in inserted:
                continue
            index = len(statements)
            statements.insert(index, feature)
            indices.append(index)

        minindex = min(indices)
        if lookups:
            feaFile.statements = statements = (
                statements[:minindex] + lookups + statements[minindex:]
            )

        # Write classDefs, anchorsDefs, markClassDefs, lookups at
        # the very top of the feature file.
        others = []
        for defs in [classDefs, anchorDefs, markClassDefs]:
            if defs:
                others.extend(defs)
                others.append(ast.Comment(""))

        feaFile.statements = statements = others + statements

    @staticmethod
    def collectInsertMarkers(feaFile, insertFeatureMarker, featureTags):
        """
        Returns a dictionary of tuples (block, comment) keyed by feature tag
        with the block that contains the comment matching the insert feature
        marker, for given feature tags.
        """
        insertComments = dict()
        for match in ast.findCommentPattern(feaFile, insertFeatureMarker):
            blocks, comment = match[:-1], match[-1]
            if len(blocks) == 1 and isinstance(blocks[0], ast.FeatureBlock):
                block = blocks[0]
                if block.name in featureTags and block.name not in insertComments:
                    insertComments[block.name] = (block, comment)
        return insertComments

    def makeUnicodeToGlyphNameMapping(self):
        """Return the Unicode to glyph name mapping for the current font."""
        # Try to get the "best" Unicode cmap subtable if this writer is running
        # in the context of a FeatureCompiler, else create a new mapping from
        # the UFO glyphs
        compiler = self.context.compiler
        cmap = None
        if compiler is not None:
            table = compiler.ttFont.get("cmap")
            if table is not None:
                cmap = table.getBestCmap()
        if cmap is None:
            from ufo2ft.util import makeUnicodeToGlyphNameMapping

            if compiler is not None:
                glyphSet = compiler.glyphSet
            else:
                glyphSet = self.context.font
            cmap = makeUnicodeToGlyphNameMapping(glyphSet)
        return cmap

    def getOrderedGlyphSet(self):
        """Return OrderedDict[glyphName, glyph] sorted by glyphOrder."""
        compiler = self.context.compiler
        if compiler is not None:
            return compiler.glyphSet

        from ufo2ft.util import _GlyphSet, makeOfficialGlyphOrder

        font = self.context.font
        # subset glyphSet by skipExportGlyphs if any
        glyphSet = _GlyphSet.from_layer(
            font,
            skipExportGlyphs=set(font.lib.get("public.skipExportGlyphs", [])),
        )
        glyphOrder = makeOfficialGlyphOrder(glyphSet, font.glyphOrder)
        return OrderedDict((gn, glyphSet[gn]) for gn in glyphOrder)

    def compileGSUB(self):
        """Compile a temporary GSUB table from the current feature file."""
        from ufo2ft.util import compileGSUB

        compiler = self.context.compiler
        fvar = None
        feafile = self.context.feaFile
        if compiler is not None:
            # The result is cached in the compiler instance, so if another
            # writer requests one it is not compiled again.
            if hasattr(compiler, "_gsub"):
                return compiler._gsub

            glyphOrder = compiler.ttFont.getGlyphOrder()
            fvar = compiler.ttFont.get("fvar")
        else:
            # the 'real' glyph order doesn't matter because the table is not
            # compiled to binary, only the glyph names are used
            glyphOrder = sorted(self.context.font.keys())

        gsub = compileGSUB(feafile, glyphOrder, fvar=fvar)

        if compiler and not hasattr(compiler, "_gsub"):
            compiler._gsub = gsub
        return gsub

    def extraSubstitutions(self):
        compiler = self.context.compiler
        if compiler is not None:
            return compiler.extraSubstitutions

    def getOpenTypeCategories(self):
        """Return 'public.openTypeCategories' values as a tuple of sets of
        unassigned, bases, ligatures, marks, components."""
        return OpenTypeCategories.load(self.context.font)

    def getGDEFGlyphClasses(self):
        """Return a tuple of GDEF GlyphClassDef base, ligature, mark, component
        glyph names.
        Sets are `None` if no 'public.openTypeCategories' values are defined or
        if no GDEF table is defined in the feature file.
        """
        feaFile = self.context.feaFile

        if ast.findTable(feaFile, "GDEF") is not None:
            return ast.getGDEFGlyphClasses(feaFile)

        unassigned, bases, ligatures, marks, components = self.getOpenTypeCategories()

        if not any((unassigned, bases, ligatures, marks, components)):
            return ast._GDEFGlyphClasses(None, None, None, None)
        return ast._GDEFGlyphClasses(
            frozenset(bases),
            frozenset(ligatures),
            frozenset(marks),
            frozenset(components),
        )

    def guessFontScripts(self):
        """Returns a set of scripts the font probably supports.

        This is done by:

        1. Looking at all defined codepoints in a font and remembering the
           script of any of the codepoints if it is associated with just one
           script. This would remember the script of U+0780 THAANA LETTER HAA
           (Thaa) but not U+061F ARABIC QUESTION MARK (multiple scripts).
        2. Adding explicitly declared `languagesystem` scripts on top.
        """
        font = self.context.font
        glyphSet = self.context.glyphSet
        feaFile = self.context.feaFile
        single_scripts = set()

        # If we're dealing with a Designspace, look at the default source.
        if hasattr(font, "findDefault"):
            font = font.findDefault().font

        # First, detect scripts from the codepoints.
        for glyph in font:
            if glyph.name not in glyphSet or glyph.unicodes is None:
                continue
            for codepoint in glyph.unicodes:
                scripts = unicodeScriptExtensions(codepoint)
                if len(scripts) == 1:
                    single_scripts.update(scripts)

        # Then, add explicitly declared languagesystems on top.
        feaScripts = ast.getScriptLanguageSystems(feaFile)
        single_scripts.update(feaScripts.keys())

        return single_scripts

    def _getAnchor(self, glyphName, anchorName, anchor=None):
        if self.context.isVariable:
            designspace = self.context.font
            x_value = VariableScalar()
            y_value = VariableScalar()
            found = False
            for source in designspace.sources:
                if source.layerName is None:
                    layer = source.font
                else:
                    layer = source.font.layers[source.layerName]
                if glyphName not in layer:
                    continue
                glyph = layer[glyphName]
                for anchor in glyph.anchors:
                    if anchor.name == anchorName:
                        location = get_userspace_location(designspace, source.location)
                        x_value.add_value(location, otRound(anchor.x))
                        y_value.add_value(location, otRound(anchor.y))
                        found = True
            if not found:
                return None
            x, y = collapse_varscalar(x_value), collapse_varscalar(y_value)
        else:
            if anchor is None:
                if glyphName not in self.context.font:
                    return None
                glyph = self.context.font[glyphName]
                anchors = [
                    anchor for anchor in glyph.anchors if anchor.name == anchorName
                ]
                if not anchors:
                    return None
                anchor = anchors[0]

            x = anchor.x
            y = anchor.y
            if hasattr(self.options, "quantization"):
                x = quantize(x, self.options.quantization)
                y = quantize(y, self.options.quantization)
        return x, y



from dataclasses import dataclass
from typing import Optional, Type

from ufo2ft.constants import CFFOptimization
from ufo2ft.outlineCompiler import OutlineOTFCompiler
from ufo2ft.preProcessor import OTFInterpolatablePreProcessor

from .interpolatableOTFCompiler import InterpolatableOTFCompiler


@dataclass
class VariableCFF2sCompiler(InterpolatableOTFCompiler):
    preProcessorClass: Type = OTFInterpolatablePreProcessor
    outlineCompilerClass: Type = OutlineOTFCompiler
    roundTolerance: Optional[float] = None
    colrAutoClipBoxes: bool = False
    cffVersion: int = 2
    optimizeCFF: CFFOptimization = CFFOptimization.SPECIALIZE
    excludeVariationTables: tuple = ()
    variableFeatures: bool = True



"""Helpers to build or extract data from feaLib AST objects."""

import collections
import functools
import operator
import re

# we re-export here all the feaLib AST classes so they can be used from
# writer modules with a single `from ufo2ft.featureWriters import ast`
import sys

from fontTools import unicodedata
from fontTools.feaLib import ast

self = sys.modules[__name__]
for name in getattr(ast, "__all__", dir(ast)):
    if isinstance(getattr(ast, name), type):
        setattr(self, name, getattr(ast, name))
del sys, self, name


def getScriptLanguageSystems(feaFile, excludeDflt=True):
    """Return dictionary keyed by Unicode script code containing lists of
    (OT_SCRIPT_TAG, [OT_LANGUAGE_TAG, ...]) tuples (excluding "DFLT" by default,
    unless excludeDflt is False).
    """
    languagesByScript = collections.OrderedDict()
    for ls in [
        st for st in feaFile.statements if isinstance(st, ast.LanguageSystemStatement)
    ]:
        if ls.script == "DFLT" and excludeDflt:
            continue
        languagesByScript.setdefault(ls.script, []).append(ls.language)

    langSysMap = collections.OrderedDict()
    for script, languages in languagesByScript.items():
        sc = unicodedata.ot_tag_to_script(script)
        langSysMap.setdefault(sc, []).append((script, languages))
    return langSysMap


def iterFeatureBlocks(feaFile, tag=None):
    for statement in feaFile.statements:
        if isinstance(statement, ast.FeatureBlock):
            if tag is not None and statement.name != tag:
                continue
            yield statement


def findFeatureTags(feaFile):
    return {f.name for f in iterFeatureBlocks(feaFile)}


def findCommentPattern(feaFile, pattern):
    """
    Yield a tuple of statements, starting with the parent block, followed by
    nested blocks if present, ending with the comment matching a given pattern.
    There is not parent block if the matched comment is a the root level.
    """
    for statement in feaFile.statements:
        if hasattr(statement, "statements"):
            for res in findCommentPattern(statement, pattern):
                yield (statement, *res)
        elif isinstance(statement, ast.Comment):
            if re.match(pattern, str(statement)):
                yield (statement,)


def findTable(feaLib, tag):
    for statement in feaLib.statements:
        if isinstance(statement, ast.TableBlock) and statement.name == tag:
            return statement


def iterClassDefinitions(feaFile):
    for s in feaFile.statements:
        if isinstance(s, ast.GlyphClassDefinition):
            yield s
        elif isinstance(s, (ast.Block)):
            yield from iterClassDefinitions(s)


LOOKUP_FLAGS = {
    "RightToLeft": 1,
    "IgnoreBaseGlyphs": 2,
    "IgnoreLigatures": 4,
    "IgnoreMarks": 8,
}


def makeLookupFlag(flags=None, markAttachment=None, markFilteringSet=None):
    if isinstance(flags, str):
        value = LOOKUP_FLAGS[flags]
    elif flags is not None:
        value = functools.reduce(operator.or_, [LOOKUP_FLAGS[n] for n in flags], 0)
    else:
        value = 0

    if markAttachment is not None:
        assert isinstance(markAttachment, ast.GlyphClassDefinition)
        markAttachment = ast.GlyphClassName(markAttachment)

    if markFilteringSet is not None:
        assert isinstance(markFilteringSet, ast.GlyphClassDefinition)
        markFilteringSet = ast.GlyphClassName(markFilteringSet)

    return ast.LookupFlagStatement(
        value, markAttachment=markAttachment, markFilteringSet=markFilteringSet
    )


def makeGlyphClassDefinitions(groups, feaFile=None, stripPrefix=""):
    """Given a groups dictionary ({str: list[str]}), create feaLib
    GlyphClassDefinition objects for each group.
    Return a dict keyed by the original group name.

    If `stripPrefix` (str) is provided and a group name starts with it,
    the string will be stripped from the beginning of the class name.
    """
    classDefs = {}
    if feaFile is not None:
        classNames = {cdef.name for cdef in iterClassDefinitions(feaFile)}
    else:
        classNames = set()
    lengthPrefix = len(stripPrefix)
    for groupName, members in sorted(groups.items()):
        originalGroupName = groupName
        if stripPrefix and groupName.startswith(stripPrefix):
            groupName = groupName[lengthPrefix:]
        className = makeFeaClassName(groupName, classNames)
        classNames.add(className)
        classDef = makeGlyphClassDefinition(className, members)
        classDefs[originalGroupName] = classDef
    return classDefs


def makeGlyphClassDefinition(className, members):
    glyphNames = [ast.GlyphName(g) for g in members]
    glyphClass = ast.GlyphClass(glyphNames)
    classDef = ast.GlyphClassDefinition(className, glyphClass)
    return classDef


def makeFeaClassName(name, existingClassNames=None):
    """Make a glyph class name which is legal to use in feature text.

    Ensures the name only includes characters in "A-Za-z0-9._", and
    isn't already defined.
    """
    name = re.sub(r"[^A-Za-z0-9._]", r"", name)
    if existingClassNames is None:
        return name
    i = 1
    origName = name
    while name in existingClassNames:
        name = "%s_%d" % (origName, i)
        i += 1
    return name


def addLookupReferences(
    feature, lookups, script=None, languages=None, exclude_dflt=False
):
    """Add references to named lookups to the feature's statements.
    If `script` (str) and `languages` (sequence of str) are provided,
    only register the lookup for the given script and languages,
    optionally with `exclude_dflt` directive.
    Otherwise add a global reference which will be registered for all
    the scripts and languages in the feature file's `languagesystems`
    statements.
    """
    assert lookups
    if not script:
        for lookup in lookups:
            feature.statements.append(ast.LookupReferenceStatement(lookup))
        return

    feature.statements.append(ast.ScriptStatement(script))
    if exclude_dflt:
        for language in languages or ("dflt",):
            feature.statements.append(
                ast.LanguageStatement(language, include_default=False)
            )
            for lookup in lookups:
                feature.statements.append(ast.LookupReferenceStatement(lookup))
    else:
        feature.statements.append(ast.LanguageStatement("dflt", include_default=True))
        for lookup in lookups:
            feature.statements.append(ast.LookupReferenceStatement(lookup))
        for language in languages or ():
            if language == "dflt":
                continue
            feature.statements.append(
                ast.LanguageStatement(language, include_default=True)
            )


_GDEFGlyphClasses = collections.namedtuple(
    "_GDEFGlyphClasses", "base ligature mark component"
)


def getGDEFGlyphClasses(feaLib):
    """Return GDEF GlyphClassDef base/mark/ligature/component glyphs, or
    None if no GDEF table is defined in the feature file.
    """
    for s in feaLib.statements:
        if isinstance(s, ast.TableBlock) and s.name == "GDEF":
            for st in s.statements:
                if isinstance(st, ast.GlyphClassDefStatement):
                    return _GDEFGlyphClasses(
                        (
                            frozenset(st.baseGlyphs.glyphSet())
                            if st.baseGlyphs is not None
                            else frozenset()
                        ),
                        (
                            frozenset(st.ligatureGlyphs.glyphSet())
                            if st.ligatureGlyphs is not None
                            else frozenset()
                        ),
                        (
                            frozenset(st.markGlyphs.glyphSet())
                            if st.markGlyphs is not None
                            else frozenset()
                        ),
                        (
                            frozenset(st.componentGlyphs.glyphSet())
                            if st.componentGlyphs is not None
                            else frozenset()
                        ),
                    )
    return _GDEFGlyphClasses(None, None, None, None)
